#ifndef __TUM3D__LARGEARRAY2D_H__
#define __TUM3D__LARGEARRAY2D_H__

// 2008-07-11 [Florian Wendel]:  added CopyTo(), CopyFrom() methods
// 2009-07-28 [Jens Schneider]:  ConvertToRaw() now writes entire lines instead of elements
// 2009-07-28 [Jens Schneider]:  Provided override for Manage Volume Privileges (Reason: Trouble on USB disks)

// TODO: Thread-Safety requires a reference counter on pages.
//       Reason: Get() method!

#include <stdio.h>
#include <windows.h>
#include <list>
#include <vector>
#include <assert.h>

#include "global.h"

struct Page;

struct Frame
{
	BYTE*						m_pData;
	Page*						m_pPage; // NULL iff free
};

struct Page
{
	Frame*						m_pFrame; // NULL iff paged out
	unsigned int				m_uiPageX;
	unsigned int				m_uiPageY;
	std::list<Page*>::iterator	m_pageCacheLocator;
	bool						m_bIsInitialized;
};

template <class T, unsigned int DIM = 1>
class LargeArray2D
{
public:
	LargeArray2D();
	~LargeArray2D();

	// Create new array. FOR USB-DRIVES UNDER WIN XP, SET bNoVolumePrivilege=false!!!
	void Create(unsigned int uiSizeX, unsigned int uiSizeY, unsigned int uiPageSizeX, unsigned int uiPageSizeY, const wchar_t* pwcFileName, unsigned __int64 ui64MemoryUsageLimit, bool bNoVolumePrivilege=false);
	// Open exisiting array
	void Open(const wchar_t* pwcFileName, bool bReadOnly, unsigned __int64 ui64MemoryUsageLimit);

	void SetMemoryUsageLimit(unsigned __int64 ui64MemoryUsageLimit);

	unsigned int GetSizeX() const { return m_uiSizeX; }
	unsigned int GetSizeY() const { return m_uiSizeY; }
	unsigned int GetPageSizeX() const { return m_uiPageSizeX; }
	unsigned int GetPageSizeY() const { return m_uiPageSizeY; }

	T* Get(unsigned int uiX, unsigned int uiY);
	void Initialize(const T* pDefaultValue);

	void CopyTo(void* pDest, unsigned int uiX, unsigned int uiY, unsigned int uiWidth, unsigned int uiHeight, unsigned int uiDestWidth = 0);
	void CopyFrom(void* pSource, unsigned int uiX, unsigned int uiY, unsigned int uiWidth, unsigned int uiHeight, unsigned int uiSourceWidth = 0);

	void ConvertToRAW(const wchar_t* pwcFileName);

	void Close();

private:
	void WriteHeader();
	void ReadHeader();
	void WriteInitFlags();
	void ReadInitFlags();

	static bool ToggleManageVolumePrivilege(bool bEnable);

	Frame* AllocateFrame();
	void DisposeFrame(Frame* pFrame);

	void PageIn(Page* pPage);
	void PageOut(Page* pPage);

	void GetPage(unsigned int uiPageX, unsigned int uiPageY);

	bool				m_bIsOpen;

	bool				m_bReadOnly;
	DWORD				m_dwSystemPageSize;

	unsigned int		m_uiSizeX, m_uiSizeY;
	unsigned int		m_uiPageSizeX, m_uiPageSizeY;
	unsigned int		m_uiNumPagesX, m_uiNumPagesY;

	unsigned int		m_uiInitFlagsMemorySize;
	unsigned int		m_uiPageMemorySize;

	T					m_defaultValue[DIM];

	HANDLE				m_hFile;

	std::vector<Frame*>	m_pFrames;
	std::vector<Frame*>	m_pFreeFrames;

	Page*				m_pages;
	std::list<Page*>	m_pageCache;

	unsigned int		m_uiLastPageX, m_uiLastPageY;
	BYTE*				m_pLastPageData;
};


template<class T, unsigned int DIM>
LargeArray2D<T, DIM>::LargeArray2D()
: m_bIsOpen(false)
{
}



template<class T, unsigned int DIM>
void LargeArray2D<T, DIM>::Create(unsigned int uiSizeX, unsigned int uiSizeY, unsigned int uiPageSizeX, unsigned int uiPageSizeY, const wchar_t* pwcFileName, unsigned __int64 ui64MemoryUsageLimit, bool bNoVolumePrivilege)
{
	assert(uiSizeX > 0);
	assert(uiSizeY > 0);
	assert(uiPageSizeX > 0);
	assert(uiPageSizeY > 0);
	
	if (m_bIsOpen)
	{
		Close();
	}

	m_bReadOnly = false;

	SYSTEM_INFO systemInfo;
	GetSystemInfo(&systemInfo);
	m_dwSystemPageSize = systemInfo.dwPageSize;

	m_uiSizeX = uiSizeX;
	m_uiSizeY = uiSizeY;
	m_uiPageSizeX = uiPageSizeX;
	m_uiPageSizeY = uiPageSizeY;
	m_uiNumPagesX = (m_uiSizeX + m_uiPageSizeX - 1) / m_uiPageSizeX;
	m_uiNumPagesY = (m_uiSizeY + m_uiPageSizeY - 1) / m_uiPageSizeY;
	
	m_uiInitFlagsMemorySize = ((sizeof(T) + sizeof(bool) * m_uiNumPagesX * m_uiNumPagesY + m_dwSystemPageSize - 1) / m_dwSystemPageSize) * m_dwSystemPageSize;
	m_uiPageMemorySize = ((DIM * sizeof(T) * m_uiPageSizeX * m_uiPageSizeY + m_dwSystemPageSize - 1) / m_dwSystemPageSize) * m_dwSystemPageSize;

	wprintf_s(L"Page memory size: %.1f MB\n", static_cast<double>(m_uiPageMemorySize) / (1024.0 * 1024.0));

	ZeroMemory(m_defaultValue, DIM * sizeof(T));

	// Create file

	// Enable Manage Volume privilege, if present in the access token associated with the process
	// This privilege is required by the SetFileValidData function which is used to avoid filling the file with zeroes
	// The privilege has to be enabled _before_ creating the file
	
	bool bHasManageVolumePrivilege = (bNoVolumePrivilege ? false : ToggleManageVolumePrivilege(true));

	if (!bHasManageVolumePrivilege)
	{
		wprintf_s(L"WARNING: Running without Manage Volume privilege\n");
	}

	m_hFile = CreateFile(pwcFileName, GENERIC_READ | GENERIC_WRITE, FILE_SHARE_READ, NULL, CREATE_ALWAYS, FILE_FLAG_NO_BUFFERING, NULL);
	if (m_hFile == INVALID_HANDLE_VALUE)
	{
		wprintf_s(L"ERROR: Cannot create file %s " __LOCATION__ L"\n", pwcFileName);
		exit(-1);
	}

	// Extend file

	LARGE_INTEGER liPos, liNewPos;
	liPos.QuadPart = static_cast<__int64>(m_dwSystemPageSize) + static_cast<__int64>(m_uiInitFlagsMemorySize) + static_cast<__int64>(m_uiPageMemorySize) * static_cast<__int64>(m_uiNumPagesX * m_uiNumPagesY);
	BOOL bResult = SetFilePointerEx(m_hFile, liPos, &liNewPos, FILE_BEGIN);
	if (!bResult || liNewPos.QuadPart != liPos.QuadPart)
	{
		wprintf_s(L"ERROR: SetFilePointerEx failed " __LOCATION__ L"\n");
		exit(-1);
	}

	bResult = SetEndOfFile(m_hFile); // Set physical end of file
	if (!bResult)
	{
		wprintf_s(L"ERROR: SetEndOfFile failed " __LOCATION__ L"\n");
		exit(-1);
	}

	if (bHasManageVolumePrivilege)
	{
		bResult = SetFileValidData(m_hFile, liPos.QuadPart); // Set logical end of file to avoid filling the file with zeroes
		if (!bResult)
		{
			DWORD dwError = GetLastError();
			wprintf_s(L"ERROR: SetFileValidData failed " __LOCATION__ L"\n", dwError);
			wprintf(  L"       If you tried to create a new file on a removable disk, such as\n");
			wprintf(  L"       an USB drive, try setting bNoVolumePrivilege = true.\n");
			wprintf(  L"       However, note that this is costly.\n");
			exit(-1);
		}
	}

	ToggleManageVolumePrivilege(false);

	WriteHeader();

	m_pages = new Page[m_uiNumPagesX * m_uiNumPagesY];
	ZeroMemory(m_pages, sizeof(Page) * m_uiNumPagesX * m_uiNumPagesY);
	for (unsigned int uiPageY = 0; uiPageY < m_uiNumPagesY; uiPageY++)
	{
		for (unsigned int uiPageX = 0; uiPageX < m_uiNumPagesX; uiPageX++)
		{
			Page* pPage = &m_pages[m_uiNumPagesX * uiPageY + uiPageX];
			pPage->m_uiPageX = uiPageX;
			pPage->m_uiPageY = uiPageY; 
		}
	}

	m_uiLastPageX = -1;
	m_uiLastPageY = -1;
	m_pLastPageData = NULL;

	SetMemoryUsageLimit(ui64MemoryUsageLimit);

	m_bIsOpen = true;

}


template<class T, unsigned int DIM>
void LargeArray2D<T, DIM>::Open(const wchar_t* pwcFileName, bool bReadOnly, unsigned __int64 ui64MemoryUsageLimit)
{
	if (m_bIsOpen)
	{
		Close();
	}

	m_bReadOnly = bReadOnly;

	SYSTEM_INFO systemInfo;
	GetSystemInfo(&systemInfo);
	m_dwSystemPageSize = systemInfo.dwPageSize;

	// Open file

	DWORD dwDesiredAccess = GENERIC_READ;
	if (!m_bReadOnly) { dwDesiredAccess |= GENERIC_WRITE; }
	m_hFile = CreateFile(pwcFileName, dwDesiredAccess, FILE_SHARE_READ, NULL, OPEN_EXISTING, FILE_FLAG_NO_BUFFERING, NULL);
	if (m_hFile == INVALID_HANDLE_VALUE)
	{
		wprintf_s(L"ERROR: Cannot open file %s " __LOCATION__ L"\n", pwcFileName);
		exit(-1);
	}

	ReadHeader();

	m_uiNumPagesX = (m_uiSizeX + m_uiPageSizeX - 1) / m_uiPageSizeX;
	m_uiNumPagesY = (m_uiSizeY + m_uiPageSizeY - 1) / m_uiPageSizeY;

	m_uiInitFlagsMemorySize = ((sizeof(T) + sizeof(bool) * m_uiNumPagesX * m_uiNumPagesY + m_dwSystemPageSize - 1) / m_dwSystemPageSize) * m_dwSystemPageSize;
	m_uiPageMemorySize = ((DIM * sizeof(T) * m_uiPageSizeX * m_uiPageSizeY + m_dwSystemPageSize - 1) / m_dwSystemPageSize) * m_dwSystemPageSize;
	
	wprintf_s(L"Page memory size: %.1f MB\n", static_cast<double>(m_uiPageMemorySize) / (1024.0 * 1024.0));

	m_pages = new Page[m_uiNumPagesX * m_uiNumPagesY];
	ZeroMemory(m_pages, sizeof(Page) * m_uiNumPagesX * m_uiNumPagesY);
	for (unsigned int uiPageY = 0; uiPageY < m_uiNumPagesY; uiPageY++)
	{
		for (unsigned int uiPageX = 0; uiPageX < m_uiNumPagesX; uiPageX++)
		{
			Page* pPage = &m_pages[m_uiNumPagesX * uiPageY + uiPageX];
			pPage->m_uiPageX = uiPageX;
			pPage->m_uiPageY = uiPageY; 
		}
	}

	ReadInitFlags();

	m_uiLastPageX = -1;
	m_uiLastPageY = -1;
	m_pLastPageData = NULL;
	
	SetMemoryUsageLimit(ui64MemoryUsageLimit);

	m_bIsOpen = true;

}


template<class T, unsigned int DIM>
void LargeArray2D<T, DIM>::WriteHeader()
{
	LARGE_INTEGER liPos, liNewPos;
	liPos.QuadPart = 0;
	BOOL bResult = SetFilePointerEx(m_hFile, liPos, &liNewPos, FILE_BEGIN);
	if (!bResult || liNewPos.QuadPart != liPos.QuadPart)
	{
		wprintf_s(L"ERROR: SetFilePointerEx failed " __LOCATION__ L"\n");
		exit(-1);
	}

	BYTE* pHeader = (BYTE*)VirtualAlloc(NULL, m_dwSystemPageSize, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE);
	if (pHeader == NULL)
	{
		wprintf_s(L"ERROR: VirtualAlloc failed " __LOCATION__ L"\n");
		exit(-1);
	}

	ZeroMemory(pHeader, m_dwSystemPageSize);
	BYTE* ptr = pHeader;
	memcpy_s(ptr, m_dwSystemPageSize, "LARGEARRAY2D", 12 * sizeof(char));
	ptr += sizeof(char) * 12;
	*reinterpret_cast<DWORD*>(ptr) = m_dwSystemPageSize;
	ptr += sizeof(DWORD);
	*reinterpret_cast<unsigned int*>(ptr) = sizeof(T);
	ptr += sizeof(unsigned int);
	*reinterpret_cast<unsigned int*>(ptr) = DIM;
	ptr += sizeof(unsigned int);
	*reinterpret_cast<unsigned int*>(ptr) = m_uiSizeX;
	ptr += sizeof(unsigned int);
	*reinterpret_cast<unsigned int*>(ptr) = m_uiSizeY;
	ptr += sizeof(unsigned int);
	*reinterpret_cast<unsigned int*>(ptr) = m_uiPageSizeX;
	ptr += sizeof(unsigned int);
	*reinterpret_cast<unsigned int*>(ptr) = m_uiPageSizeY;
	ptr += sizeof(unsigned int);

	DWORD dwNumberOfBytesWritten;
	bResult = WriteFile(m_hFile, pHeader, m_dwSystemPageSize, &dwNumberOfBytesWritten, NULL);
	if (!bResult || dwNumberOfBytesWritten != m_dwSystemPageSize)
	{
		wprintf_s(L"ERROR: WriteFile failed " __LOCATION__ L"\n");
		exit(-1);
	}

	bResult = VirtualFree(pHeader, 0, MEM_RELEASE);
	if (!bResult)
	{
		wprintf_s(L"ERROR: VirtualFree failed " __LOCATION__ L"\n");
		exit(-1);
	}
}


template<class T, unsigned int DIM>
void LargeArray2D<T, DIM>::ReadHeader()
{
	LARGE_INTEGER liPos, liNewPos;
	liPos.QuadPart = 0;
	BOOL bResult = SetFilePointerEx(m_hFile, liPos, &liNewPos, FILE_BEGIN);
	if (!bResult || liNewPos.QuadPart != liPos.QuadPart)
	{
		wprintf_s(L"ERROR: SetFilePointerEx failed " __LOCATION__ L"\n");
		exit(-1);
	}

	BYTE* pHeader = (BYTE*)VirtualAlloc(NULL, m_dwSystemPageSize, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE);
	if (pHeader == NULL)
	{
		wprintf_s(L"ERROR: VirtualAlloc failed " __LOCATION__ L"\n");
		exit(-1);
	}

	DWORD dwNumberOfBytesRead;
	bResult = ReadFile(m_hFile, pHeader, m_dwSystemPageSize, &dwNumberOfBytesRead, NULL);
	if (!bResult || dwNumberOfBytesRead != m_dwSystemPageSize)
	{
		wprintf_s(L"ERROR: ReadFile failed " __LOCATION__ L"\n");
		exit(-1);
	}

	BYTE* ptr = pHeader;
	if (strncmp(reinterpret_cast<char*>(ptr), "LARGEARRAY2D", 12) != 0)
	{
		wprintf_s(L"ERROR: Signature check failed " __LOCATION__ L"\n");
		exit(-1);
	}
	ptr += sizeof(char) * 12;
	if (*reinterpret_cast<DWORD*>(ptr) != m_dwSystemPageSize)
	{
		wprintf_s(L"ERROR: System page size check failed " __LOCATION__ L"\n");
		exit(-1);
	}
	ptr += sizeof(DWORD);
	if (*reinterpret_cast<unsigned int*>(ptr) != sizeof(T))
	{
		wprintf_s(L"ERROR: Type size check failed " __LOCATION__ L"\n");
		exit(-1);
	}
	ptr += sizeof(unsigned int);
	if (*reinterpret_cast<unsigned int*>(ptr) != DIM)
	{
		wprintf_s(L"ERROR: Dimension check failed " __LOCATION__ L"\n");
		exit(-1);
	}
	ptr += sizeof(unsigned int);
	m_uiSizeX = *reinterpret_cast<unsigned int*>(ptr);
	ptr += sizeof(unsigned int);
	m_uiSizeY = *reinterpret_cast<unsigned int*>(ptr);
	ptr += sizeof(unsigned int);
	m_uiPageSizeX = *reinterpret_cast<unsigned int*>(ptr);
	ptr += sizeof(unsigned int);
	m_uiPageSizeY = *reinterpret_cast<unsigned int*>(ptr);
	ptr += sizeof(unsigned int);

	bResult = VirtualFree(pHeader, 0, MEM_RELEASE);
	if (!bResult)
	{
		wprintf_s(L"ERROR: VirtualFree failed " __LOCATION__ L"\n");
		exit(-1);
	}
}


template<class T, unsigned int DIM>
void LargeArray2D<T, DIM>::WriteInitFlags()
{
	LARGE_INTEGER liPos, liNewPos;
	liPos.QuadPart = m_dwSystemPageSize;
	BOOL bResult = SetFilePointerEx(m_hFile, liPos, &liNewPos, FILE_BEGIN);
	if (!bResult || liNewPos.QuadPart != liPos.QuadPart)
	{
		wprintf_s(L"ERROR: SetFilePointerEx failed " __LOCATION__ L"\n");
		exit(-1);
	}

	BYTE* pInitFlags = (BYTE*)VirtualAlloc(NULL, m_uiInitFlagsMemorySize, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE);
	if (pInitFlags == NULL)
	{
		wprintf_s(L"ERROR: VirtualAlloc failed " __LOCATION__ L"\n");
		exit(-1);
	}

	ZeroMemory(pInitFlags, m_uiInitFlagsMemorySize);
	BYTE* ptr = pInitFlags;
	for (unsigned int ui = 0; ui < DIM; ui++)
	{
		*reinterpret_cast<T*>(ptr) = m_defaultValue[ui];
		ptr += sizeof(T);
	}
	for (unsigned int uiPage = 0; uiPage < m_uiNumPagesX * m_uiNumPagesY; uiPage++)
	{
		*reinterpret_cast<bool*>(ptr) = m_pages[uiPage].m_bIsInitialized;
		ptr += sizeof(bool);
	}

	DWORD dwNumberOfBytesWritten;
	bResult = WriteFile(m_hFile, pInitFlags, m_uiInitFlagsMemorySize, &dwNumberOfBytesWritten, NULL);
	if (!bResult || dwNumberOfBytesWritten != m_uiInitFlagsMemorySize)
	{
		wprintf_s(L"ERROR: WriteFile failed " __LOCATION__ L"\n");
		exit(-1);
	}

	bResult = VirtualFree(pInitFlags, 0, MEM_RELEASE);
	if (!bResult)
	{
		wprintf_s(L"ERROR: VirtualFree failed " __LOCATION__ L"\n");
		exit(-1);
	}
}


template<class T, unsigned int DIM>
void LargeArray2D<T, DIM>::ReadInitFlags()
{
	LARGE_INTEGER liPos, liNewPos;
	liPos.QuadPart = m_dwSystemPageSize;
	BOOL bResult = SetFilePointerEx(m_hFile, liPos, &liNewPos, FILE_BEGIN);
	if (!bResult || liNewPos.QuadPart != liPos.QuadPart)
	{
		wprintf_s(L"ERROR: SetFilePointerEx failed " __LOCATION__ L"\n");
		exit(-1);
	}

	BYTE* pInitFlags = (BYTE*)VirtualAlloc(NULL, m_uiInitFlagsMemorySize, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE);
	if (pInitFlags == NULL)
	{
		wprintf_s(L"ERROR: VirtualAlloc failed " __LOCATION__ L"\n");
		exit(-1);
	}

	DWORD dwNumberOfBytesRead;
	bResult = ReadFile(m_hFile, pInitFlags, m_uiInitFlagsMemorySize, &dwNumberOfBytesRead, NULL);
	if (!bResult || dwNumberOfBytesRead != m_uiInitFlagsMemorySize)
	{
		wprintf_s(L"ERROR: ReadFile failed " __LOCATION__ L"\n");
		exit(-1);
	}

	BYTE* ptr = pInitFlags;
	for (unsigned int ui = 0; ui < DIM; ui++)
	{
		m_defaultValue[ui] = *reinterpret_cast<T*>(ptr);
		ptr += sizeof(T);
	}
	for (unsigned int uiPage = 0; uiPage < m_uiNumPagesX * m_uiNumPagesY; uiPage++)
	{
		m_pages[uiPage].m_bIsInitialized = *reinterpret_cast<bool*>(ptr);
		ptr += sizeof(bool);
	}

	bResult = VirtualFree(pInitFlags, 0, MEM_RELEASE);
	if (!bResult)
	{
		wprintf_s(L"ERROR: VirtualFree failed " __LOCATION__ L"\n");
		exit(-1);
	}
}


template<class T, unsigned int DIM>
bool LargeArray2D<T, DIM>::ToggleManageVolumePrivilege(bool bEnable)
{
	HANDLE hToken;
	BOOL bResult = OpenProcessToken(GetCurrentProcess(), TOKEN_ALL_ACCESS, &hToken);
	if (!bResult)
	{
		wprintf_s(L"ERROR: OpenProcessToken failed " __LOCATION__ L"\n");
		exit(-1);
	}

	LUID luid;
	bResult = LookupPrivilegeValue(NULL, SE_MANAGE_VOLUME_NAME, &luid);
	if (!bResult)
	{
		wprintf_s(L"ERROR: LookupPrivilegeValue failed " __LOCATION__ L"\n");
		exit(-1);
	}

	TOKEN_PRIVILEGES tp;
	tp.PrivilegeCount = 1;
	tp.Privileges[0].Luid = luid;
	tp.Privileges[0].Attributes = bEnable ? SE_PRIVILEGE_ENABLED : 0;
	bResult = AdjustTokenPrivileges(hToken, FALSE, &tp, sizeof(TOKEN_PRIVILEGES), NULL, NULL);
	if (!bResult)
	{
		wprintf_s(L"ERROR: AdjustTokenPrivileges failed " __LOCATION__ L"\n");
		exit(-1);
	}

	bool bHasManageVolumePrivilege = GetLastError() != ERROR_NOT_ALL_ASSIGNED;

	bResult = CloseHandle(hToken);
	if (!bResult)
	{
		wprintf_s(L"ERROR: CloseHandle failed " __LOCATION__ L"\n");
		exit(-1);
	}

	return bHasManageVolumePrivilege;
}


template<class T, unsigned int DIM>
void LargeArray2D<T, DIM>::SetMemoryUsageLimit(unsigned __int64 ui64MemoryUsageLimit)
{
	unsigned int uiNumFrames = max(static_cast<unsigned int>(ui64MemoryUsageLimit / static_cast<unsigned __int64>(m_uiPageMemorySize)), 1);

	if (uiNumFrames < m_pFrames.size())
	{
		unsigned int uiNumFramesToDispose = static_cast<unsigned int>(m_pFrames.size()) - uiNumFrames;
		for (unsigned int ui = 0; ui < uiNumFramesToDispose; ui++)
		{
			Frame* pFrame = m_pFrames.back();
			m_pFrames.pop_back();
			if (pFrame->m_pPage != NULL)
			{
				PageOut(pFrame->m_pPage);
			}
			for (std::vector<Frame*>::iterator p = m_pFreeFrames.begin(); p != m_pFreeFrames.end(); p++)
			{
				if (*p == pFrame)
				{
					m_pFreeFrames.erase(p);
					break;
				}
			}
			DisposeFrame(pFrame);
		}
		m_uiLastPageX = -1;
		m_uiLastPageY = -1;
		m_pLastPageData = NULL;
	}
	else
	{
		unsigned int uiNumFramesToAllocate = uiNumFrames - static_cast<unsigned int>(m_pFrames.size());
		for (unsigned int ui = 0; ui < uiNumFramesToAllocate; ui++)
		{
			Frame* pFrame = AllocateFrame();
			m_pFrames.push_back(pFrame);
			m_pFreeFrames.push_back(pFrame);
		}
	}

	wprintf_s(L"Number of frames allocated: %u\n", uiNumFrames);
}


template<class T, unsigned int DIM>
Frame* LargeArray2D<T, DIM>::AllocateFrame()
{
	Frame* pFrame = new Frame();
	ZeroMemory(pFrame, sizeof(Frame));

	pFrame->m_pData = (BYTE*)VirtualAlloc(NULL, m_uiPageMemorySize, MEM_RESERVE | MEM_COMMIT, PAGE_READWRITE);
	if (pFrame->m_pData == NULL)
	{
		wprintf_s(L"ERROR: VirtualAlloc failed " __LOCATION__ L"\n");
		exit(-1);
	}
	ZeroMemory(pFrame->m_pData, m_uiPageMemorySize);

	return pFrame;
}


template<class T, unsigned int DIM>
void LargeArray2D<T, DIM>::DisposeFrame(Frame* pFrame)
{
	BOOL bResult = VirtualFree(pFrame->m_pData, 0, MEM_RELEASE);
	if (!bResult)
	{
		wprintf_s(L"ERROR: VirtualFree failed " __LOCATION__ L"\n");
		exit(-1);
	}

	delete pFrame;
}


template<class T, unsigned int DIM>
void LargeArray2D<T, DIM>::PageIn(Page* pPage)
{
	assert(pPage->m_pFrame == NULL);
	assert(m_pFreeFrames.size() > 0);

	Frame* pFrame = m_pFreeFrames.back();
	m_pFreeFrames.pop_back();

	pFrame->m_pPage = pPage;
	pPage->m_pFrame = pFrame;

	if (pPage->m_bIsInitialized)
	{
		LARGE_INTEGER liPos, liNewPos;
		liPos.QuadPart = static_cast<__int64>(m_dwSystemPageSize) + static_cast<__int64>(m_uiInitFlagsMemorySize) + static_cast<__int64>(m_uiPageMemorySize) * static_cast<__int64>(m_uiNumPagesX * pPage->m_uiPageY + pPage->m_uiPageX);
		BOOL bResult = SetFilePointerEx(m_hFile, liPos, &liNewPos, FILE_BEGIN);
		if (!bResult || liNewPos.QuadPart != liPos.QuadPart)
		{
			wprintf_s(L"ERROR: SetFilePointerEx failed " __LOCATION__ L"\n");
			exit(-1);
		}
		
		DWORD dwNumBytesRead;
		bResult = ReadFile(m_hFile, pPage->m_pFrame->m_pData, m_uiPageMemorySize, &dwNumBytesRead, NULL);
		if (!bResult || dwNumBytesRead != m_uiPageMemorySize)
		{
			wprintf_s(L"ERROR: ReadFile failed " __LOCATION__ L"\n");
			exit(-1);
		}

		//wprintf_s(L"Page (%u, %u) read\n", pPage->m_uiPageX, pPage->m_uiPageY);
	}
	else
	{
		BYTE* pData = pFrame->m_pData;
		for (unsigned int ui = 0; ui < m_uiPageSizeX * m_uiPageSizeY; ui++)
		{
			for (unsigned int uj = 0; uj < DIM; uj++)
			{
				reinterpret_cast<T*>(pData)[DIM * ui + uj] = m_defaultValue[uj];
			}
		}

		if (!m_bReadOnly)
		{
			pPage->m_bIsInitialized = true;
		}

		//wprintf_s(L"Page (%u, %u) initialized\n", pPage->m_uiPageX, pPage->m_uiPageY);
	}

}


template<class T, unsigned int DIM>
void LargeArray2D<T, DIM>::PageOut(Page* pPage)
{
	assert(pPage->m_pFrame != NULL);

	if (!m_bReadOnly)
	{
		LARGE_INTEGER liPos, liNewPos;
		liPos.QuadPart = static_cast<__int64>(m_dwSystemPageSize) + static_cast<__int64>(m_uiInitFlagsMemorySize) + static_cast<__int64>(m_uiPageMemorySize) * static_cast<__int64>(m_uiNumPagesX * pPage->m_uiPageY + pPage->m_uiPageX);
		BOOL bResult = SetFilePointerEx(m_hFile, liPos, &liNewPos, FILE_BEGIN);
		if (!bResult || liNewPos.QuadPart != liPos.QuadPart)
		{
			wprintf_s(L"ERROR: SetFilePointerEx failed " __LOCATION__ L"\n");
			exit(-1);
		}
		
		DWORD dwNumBytesWritten;
		bResult = WriteFile(m_hFile, pPage->m_pFrame->m_pData, m_uiPageMemorySize, &dwNumBytesWritten, NULL);
		if (!bResult || dwNumBytesWritten != m_uiPageMemorySize)
		{
			wprintf_s(L"ERROR: WriteFile failed " __LOCATION__ L"\n");
			exit(-1);
		}

		//wprintf_s(L"Page (%u, %u) written\n", pPage->m_uiPageX, pPage->m_uiPageY);
	}

	Frame* pFrame = pPage->m_pFrame;
	pPage->m_pFrame = NULL;
	pFrame->m_pPage = NULL;

	m_pFreeFrames.push_back(pFrame);
}


template<class T, unsigned int DIM>
inline T* LargeArray2D<T, DIM>::Get(unsigned int uiX, unsigned int uiY)
{
	assert(uiX < m_uiSizeX);
	assert(uiY < m_uiSizeY);

	unsigned int uiPageX = uiX / m_uiPageSizeX;
	unsigned int uiPageY = uiY / m_uiPageSizeY;
	unsigned int uiRelX = uiX % m_uiPageSizeX;
	unsigned int uiRelY = uiY % m_uiPageSizeY;

	if (uiPageX != m_uiLastPageX || uiPageY != m_uiLastPageY)
	{
		GetPage(uiPageX, uiPageY);
	}

	return &reinterpret_cast<T*>(m_pLastPageData)[DIM * (m_uiPageSizeX * uiRelY + uiRelX)];
}


template<class T, unsigned int DIM>
void LargeArray2D<T, DIM>::GetPage(unsigned int uiPageX, unsigned int uiPageY)
{
	Page* pPage = &m_pages[m_uiNumPagesX * uiPageY + uiPageX];
	if (pPage->m_pFrame == NULL)
	{
		if (m_pFreeFrames.size() == 0)
		{
			Page* pPageToPageOut = m_pageCache.back();
			m_pageCache.pop_back();
			PageOut(pPageToPageOut);
		}
		PageIn(pPage);
	}
	else
	{
		m_pageCache.erase(pPage->m_pageCacheLocator);
	}

	m_pageCache.push_front(pPage);
	pPage->m_pageCacheLocator = m_pageCache.begin();

	m_uiLastPageX = uiPageX;
	m_uiLastPageY = uiPageY;
	m_pLastPageData = pPage->m_pFrame->m_pData;
}


template<class T, unsigned int DIM>
void LargeArray2D<T, DIM>::Initialize(const T* pDefaultValue)
{
	if (!m_bReadOnly)
	{
		m_uiLastPageX = -1;
		m_uiLastPageY = -1;
		m_pLastPageData = NULL;

		m_pFreeFrames.clear();
		for (unsigned int ui = 0; ui < m_pFrames.size(); ui++)
		{
			Frame* pFrame = m_pFrames[ui];
			if (pFrame->m_pPage != NULL)
			{
				pFrame->m_pPage->m_pFrame = NULL;
				pFrame->m_pPage = NULL;
			}
			m_pFreeFrames.push_back(pFrame);
		}

		for (unsigned int ui = 0; ui < DIM; ui++)
		{
			m_defaultValue[ui] = pDefaultValue[ui];
		}
		for (unsigned int uiPage = 0; uiPage < m_uiNumPagesX * m_uiNumPagesY; uiPage++)
		{
			m_pages[uiPage].m_bIsInitialized = false;
		}
	}
}


template<class T, unsigned int DIM>
void LargeArray2D<T, DIM>::CopyTo(void* pDest, unsigned int uiX, unsigned int uiY, unsigned int uiWidth, unsigned int uiHeight, unsigned int uiDestWidth)
{
	assert(uiX < m_uiSizeX);
	assert(uiY < m_uiSizeY);
	assert(uiWidth >= 0);
	assert(uiHeight >= 0);
	assert(uiX + uiWidth <= m_uiSizeX);
	assert(uiY + uiHeight <= m_uiSizeY);
	assert(pDest != NULL);

	if (uiWidth == 0 || uiHeight == 0)
		return;
	if (uiDestWidth == 0)
		uiDestWidth = uiWidth;

	unsigned int uiFirstColumn = uiX / m_uiPageSizeX;
	unsigned int uiLastColumn  = (uiX + uiWidth - 1) / m_uiPageSizeX;
	unsigned char* pucPtr = (unsigned char*)pDest;
	unsigned char* pucDest;
	unsigned char* pucSource;

	for (unsigned int uiColumn = uiFirstColumn; uiColumn <= uiLastColumn; ++uiColumn)
	{
		unsigned int uiPageStart = uiColumn * m_uiPageSizeX;
		unsigned int uiStart = max(uiX, uiPageStart);
		unsigned int uiEnd   = min(uiX + uiWidth, uiPageStart + m_uiPageSizeX);

		for (unsigned int uiLine = 0; uiLine < uiHeight; ++uiLine)
		{
			pucDest   = &pucPtr[(uiStart - uiX + uiLine * uiDestWidth) * sizeof(T) * DIM];
			pucSource = Get(uiStart, uiY + uiLine);
			CopyMemory(pucDest, pucSource, (uiEnd - uiStart) * sizeof(T) * DIM);
		}
	}
}


template<class T, unsigned int DIM>
void LargeArray2D<T, DIM>::CopyFrom(void* pSource, unsigned int uiX, unsigned int uiY, unsigned int uiWidth, unsigned int uiHeight, unsigned int uiSourceWidth)
{
	assert(pSource != NULL);
	assert(uiX < m_uiSizeX);
	assert(uiY < m_uiSizeY);
	if (uiX + uiWidth  > m_uiSizeX)  uiWidth  = m_uiSizeX - uiX;
	if (uiY + uiHeight > m_uiSizeY)  uiHeight = m_uiSizeY - uiY;
	assert(uiWidth >= 0);
	assert(uiHeight >= 0);

	if (uiWidth == 0 || uiHeight == 0)
		return;
	if (uiSourceWidth == 0)
		uiSourceWidth = uiWidth;

	unsigned int uiFirstColumn = uiX / m_uiPageSizeX;
	unsigned int uiLastColumn  = (uiX + uiWidth - 1) / m_uiPageSizeX;
	unsigned char* pucPtr = (unsigned char*)pSource;
	unsigned char* pucDest;
	unsigned char* pucSource;

	for (unsigned int uiColumn = uiFirstColumn; uiColumn <= uiLastColumn; ++uiColumn)
	{
		unsigned int uiPageStart = uiColumn * m_uiPageSizeX;
		unsigned int uiStart = max(uiX, uiPageStart);
		unsigned int uiEnd   = min(uiX + uiWidth, uiPageStart + m_uiPageSizeX);

		for (unsigned int uiLine = 0; uiLine < uiHeight; ++uiLine)
		{
			pucDest   = Get(uiStart, uiY + uiLine);
			pucSource = &pucPtr[(uiStart - uiX + uiLine * uiSourceWidth) * sizeof(T) * DIM];
			CopyMemory(pucDest, pucSource, (uiEnd - uiStart) * sizeof(T) * DIM);
		}
	}
}


template<class T, unsigned int DIM>
void LargeArray2D<T, DIM>::ConvertToRAW(const wchar_t* pwcFileName)
{
	FILE* hFile;
	errno_t err = _wfopen_s(&hFile, pwcFileName, L"wb");
	if (err != 0)
	{
		wprintf_s(L"ERROR: Cannot open file %s " __LOCATION__ L"\n", pwcFileName);
		exit(-1);
	}
	for (unsigned int uiY = 0; uiY < m_uiSizeY; uiY++)
	{
		// perform line-wise output to file
		unsigned int uiX=0;
		while (uiX<m_uiSizeX) {
			unsigned int numElems = min(m_uiPageSizeX,m_uiSizeX-uiX);
			fwrite(Get(uiX,uiY),sizeof(T)*numElems, DIM, hFile);
			uiX+=numElems;
		}

		//for (unsigned int uiX = 0; uiX < m_uiSizeX; uiX++)
		//{
		//	fwrite(Get(uiX, uiY), sizeof(T), DIM, hFile);
		//}
	}
	fclose(hFile);
}


template<class T, unsigned int DIM>
void LargeArray2D<T, DIM>::Close()
{
	if (m_bIsOpen)
	{
		if (!m_bReadOnly)
		{
			WriteInitFlags();
		}

		for (unsigned int ui = 0; ui < m_pFrames.size(); ui++)
		{
			Frame* pFrame = m_pFrames[ui];
			if (pFrame->m_pPage != NULL)
			{
				PageOut(pFrame->m_pPage);
			}
			DisposeFrame(pFrame);
		}
		
		m_pFreeFrames.clear();
		m_pFrames.clear();
		m_pageCache.clear();

		delete[] m_pages;

		BOOL bResult = CloseHandle(m_hFile);
		if (!bResult)
		{
			wprintf_s(L"ERROR: CloseHandle failed " __LOCATION__ L"\n");
			exit(-1);
		}

		m_bIsOpen = false;
	}
}


template<class T, unsigned int DIM>
LargeArray2D<T, DIM>::~LargeArray2D()
{
	if (m_bIsOpen)
	{
		Close();
	}
}


#endif






/*

#include "largearray2d.h"

#include <stdlib.h>
#include <time.h>
#include <windows.h>


struct Point
{
	unsigned int	m_uiX;
	unsigned int	m_uiY;
	float			m_fValue1;
	float			m_fValue2;
};

int wmain(int argc, wchar_t* argv[])
{
	LargeArray2D<float,2> la;
	la.Create(20000, 20000, 1024, 1024, L"C:\\Temp\\data.la2d", 3 * 8 * 1024 * 1024);
	wprintf_s(L"Size X: %u, Size Y: %u\n", la.GetSizeX(), la.GetSizeY());

	float init[2] = {123.0f, 456.0f};
	la.Initialize(init);

	srand( (unsigned)time( NULL ) );

	const unsigned int uiNUM_POINTS = 10;
	
	Point* points = new Point[uiNUM_POINTS];
	ZeroMemory(points, uiNUM_POINTS * sizeof(Point));

	for (unsigned int ui = 0; ui < uiNUM_POINTS; ui++)
	{
		points[ui].m_uiX = (unsigned int)((double)rand() / (double)RAND_MAX * 19999.0);
		points[ui].m_uiY = (unsigned int)((double)rand() / (double)RAND_MAX * 19999.0);
		points[ui].m_fValue1 = (float)rand();
		points[ui].m_fValue2 = (float)rand();

		float* pf = la.Get(points[ui].m_uiX, points[ui].m_uiY);
		pf[0] = points[ui].m_fValue1;
		pf[1] = points[ui].m_fValue2;
	}

	la.Close();

	la.Open(L"C:\\Temp\\data.la2d", true, 3 * 8 * 1024 * 1024);
	wprintf_s(L"Size X: %u, Size Y: %u\n", la.GetSizeX(), la.GetSizeY());

	for (unsigned int ui = 0; ui < uiNUM_POINTS; ui++)
	{
		float* pf = la.Get(points[ui].m_uiX, points[ui].m_uiY);
		wprintf_s(L"Value1: %f Expected: %f Value2: %f Exptected: %f\n", pf[0], points[ui].m_fValue1, pf[1], points[ui].m_fValue2);
		if (pf[0] != points[ui].m_fValue1 || pf[1] != points[ui].m_fValue2)
		{
			wprintf_s(L"ERROR!\n");
			exit(-1);
		}
	}

	for (unsigned int ui = 0; ui < uiNUM_POINTS; ui++)
	{
		unsigned int uiX = (unsigned int)((double)rand() / (double)RAND_MAX * 19999.0);
		unsigned int uiY = (unsigned int)((double)rand() / (double)RAND_MAX * 19999.0);

		float* pf = la.Get(uiX, uiY);

		wprintf_s(L"Random Read: %f %f\n", pf[0], pf[1]);
	}

	la.Close();

}

*/
