#include <windows.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <vector>
#include <process.h>
#include "timing.h"
#include "largearray2d.h"

//#define TUM3D_WRITE_RAW_FILE	// copies LargeArray2D into a raw rgb8 file.
#define RESX		512							// resolution in X-direction in Kpixels
#define RESY		512							// resolution in Y-direction in Kpixels
#define OUTPUT		L"f:\\Mandelbrot512_L%.3i.bin"	// output file
#define MEM_USAGE	512							// Memory usage in MB for paging
//#define PYRAMID_ONLY

// TODO: generate mip-hierarchy
// TODO: compress to Gigapixel Viewer

// DONE: Large Palettes
// DONE: Anti-Aliasing
// DONE: Threading
// DONE: store in large array

using namespace std;

// minimalistic complex numbers
class COMPLEX {
public:
	COMPLEX(double _r=0.0,double _i=0.0) : r(_r), i(_i) {}
	COMPLEX(const COMPLEX& other) {
		r=other.r;
		i=other.i;
	}
	COMPLEX& operator=(const COMPLEX& other) {
		r=other.r;
		i=other.i;
		return *this;
	}
	// mandelbrot iteration
	inline void mb_iterate(const COMPLEX& c) {
		double t = r;
		r = r*r - i*i + c.r;
		i = 2*t*i + c.i;
	}
	inline COMPLEX operator+(const COMPLEX& other) const {
		return COMPLEX(r+other.r,i+other.i);
	}
	// squared modulus
	inline double modulus(void) const {
		return r*r + i*i;
	}
	double r,i;
};

// DESCRIPTOR FOR A MANDELBROT TILE
struct JOBDESC {
	COMPLEX					ll;
	COMPLEX					ur;
	size_t					width;
	size_t					height;
	size_t					numThreads;
	unsigned short			max_iter;
	unsigned char*			rgb;		// width*height*3 unsigned chars
	unsigned short*			scratch;	// (width+1)*(height+1)*2 unsigned shorts
	double					escapeSq;
	void allocateBuffers(void) {
		rgb     = new unsigned char[3*width*height];
		scratch = new unsigned short[2*(width+1)*(height+1)];
	}
	void freeBuffers(void) {
		delete[] rgb;
		delete[] scratch;
	}
	vector<unsigned char>*  LUT;
};

// DESCRIPTOR FOR A PARALLEL MANDELBROT BATCH
struct PARALLEL_JOB {
	~PARALLEL_JOB(void) {
		vIO.clear();
	}
	const JOBDESC* job;
	vector<pair<COMPLEX,unsigned short*> >	vIO;
	bool bFinished;
};

// Check wether c is in main cardioid of Mandelbrot set
// The main cardioid completely belongs to the set.
inline bool IsInMainCardioid(const COMPLEX& c) {
	COMPLEX cp(1.0-4.0*c.r,-4.0*c.i);
	// polar form of cp
	double rl=sqrt(cp.modulus());
	double phi=atan2(cp.i,cp.r);
	// find roots r1,r2 of cp
	double rsq=sqrt(rl);
	double cs=rsq*cos(phi/2.0);
	double sn=rsq*sin(phi/2.0);
	double c1=(cs+1.0)*(cs+1.0)+sn*sn;
	if (c1<1.0) return true;
	c1-=4.0*cs;
	if (c1<1.0) return true;
	return false;
}

// Check wether c is in attracting cycle of period 2
inline bool IsInCyclePeriod2(const COMPLEX& c) {
	double R=(c.r+1.0)*(c.r+1.0)+c.i*c.i;
	return (R<1/16.0);
}

inline void MandelbrotIteration(unsigned short* result, const COMPLEX& c, const JOBDESC& job) {
	COMPLEX z(0.0,0.0);

	if (IsInCyclePeriod2(c)) {
		*result = job.max_iter;
		return;
	}
	if (IsInMainCardioid(c)) {
		*result = job.max_iter;
		return;
	}
	
	// otherwise iteration
	unsigned short iter=0;
	do {
		z.mb_iterate(c);
		++iter;
	} while (z.modulus()<=job.escapeSq && iter<=job.max_iter);
	*result = iter-1;
}

// This routine computes a Mandelbrot fractal using Quad flip supersampling
void MandelbrotQuadFlipped(const JOBDESC& job) {
	COMPLEX p,c1,c2;
	double dy=1.0/static_cast<double>(job.height)*(job.ur.i-job.ll.i);
	double dx=1.0/static_cast<double>(job.height)*(job.ur.i-job.ll.i);
	// Compute Quad flipped samples
	for (size_t y=0; y<job.height+1; y++) {		
		p.i=job.ll.i+static_cast<double>(y)*dy;
		for (size_t x=0; x<job.width+1; x++) {
			p.r=job.ll.r+static_cast<double>(x)*dx;
			if ((x&1)^(y&1)) {  // odd
				c1=p+COMPLEX(-dx/2.0, dy/6.0);
				c2=p+COMPLEX( dx/6.0, dy/2.0);
			}
			else {				// even
				c1=p+COMPLEX(-dx/2.0,-dy/6.0);
				c2=p+COMPLEX(-dx/6.0, dy/2.0);
			}
			MandelbrotIteration(&job.scratch[2*(x+y*(job.width+1))  ],c1,job);
			MandelbrotIteration(&job.scratch[2*(x+y*(job.width+1))+1],c2,job);
		}
	}
	size_t LUTentries=job.LUT->size()/3;
	for (size_t y=0; y<job.height; y++) {
		for (size_t x=0; x<job.width; x++) {
			double d=static_cast<double>(job.scratch[2*(x+y*(job.width+1))  ] + 
										 job.scratch[2*(x+y*(job.width+1))+1] + 
										 job.scratch[2*((x+1)+y*(job.width+1))] + 
										 job.scratch[2*(x+(y+1)*(job.width+1))+1])/4.0;
			if (d==static_cast<double>(job.max_iter)) {
				for (int i=0; i<3; i++) job.rgb[3*(x+y*job.width)+i]=0;
			} 
			else {
				int id=static_cast<int>(d)%LUTentries;
				double w=d-floor(d);				
				for (int i=0; i<3; i++) {
					double c = static_cast<double>((*job.LUT)[3*id+i])+w*(static_cast<double>((*job.LUT)[3*((id+1)%LUTentries)+i])-static_cast<double>((*job.LUT)[3*id+i]));
					job.rgb[3*(x+y*job.width)+i]=static_cast<unsigned char>(max(min(c,255.0),0.0));
				}
			}
		}
	}
}

// THREAD-CALLBACK FUNCTION FOR ITERATOR THREADS
void __cdecl ThreadedMandelbrotIterator(void* data) {
	PARALLEL_JOB* J=reinterpret_cast<PARALLEL_JOB*>(data);
	for (vector<pair<COMPLEX,unsigned short*> >::iterator it=J->vIO.begin(); it!=J->vIO.end(); it++) {
		MandelbrotIteration(it->second,it->first,*(J->job));
	}
	J->bFinished=true;
	_endthread();
}

struct ITERATOR_THREAD {
	uintptr_t ptr;
	unsigned thrdaddr;
	PARALLEL_JOB *J;
	bool Run(int i=-1) {
		J->bFinished = false;
		ptr = _beginthread(ThreadedMandelbrotIterator,0,J);
		DWORD_PTR p=0xFF;
		if (i>-1) p = SetThreadAffinityMask(reinterpret_cast<HANDLE>(ptr),1<<i);
		if (!p || !ptr) printf("Warning: problems spawning thread!\n");
		return (ptr!=NULL);
	}
	bool IsFinished(void) {
		return J->bFinished;
	}
};

void ParallelMandelbrot(const JOBDESC& job) {
	// Set up threading -- SERIAL
	vector<ITERATOR_THREAD> vThreads;
	vThreads.resize(job.numThreads);
	for (size_t t=0; t<job.numThreads; t++) {
		vThreads[t].J = new PARALLEL_JOB();
		vThreads[t].J->job = &job;
		vThreads[t].J->vIO.reserve(2*(job.height+1)*(job.width+1)/job.numThreads+1);
	}
	
	// Distribute points among the workers -- SERIAL
	size_t t=0;
	COMPLEX p,c1,c2;
	double dy=1.0/static_cast<double>(job.height)*(job.ur.i-job.ll.i);
	double dx=1.0/static_cast<double>(job.height)*(job.ur.i-job.ll.i);
	for (size_t y=0; y<job.height+1; y++) {		
		p.i=job.ll.i+static_cast<double>(y)*dy;
		for (size_t x=0; x<job.width+1; x++) {
			p.r=job.ll.r+static_cast<double>(x)*dx;
			// Quad-Flipped Samples
			if ((x&1)^(y&1)) {  // odd
				c1=p+COMPLEX(-dx/2.0, dy/6.0);
				c2=p+COMPLEX( dx/6.0, dy/2.0);
			}
			else {				// even
				c1=p+COMPLEX(-dx/2.0,-dy/6.0);
				c2=p+COMPLEX(-dx/6.0, dy/2.0);
			}
			// Spin to threads
			vThreads[t].J->vIO.push_back(pair<COMPLEX,unsigned short*>(c1,&job.scratch[2*(x+y*(job.width+1))  ]));
			vThreads[t].J->vIO.push_back(pair<COMPLEX,unsigned short*>(c2,&job.scratch[2*(x+y*(job.width+1))+1]));
			t=(t+1)%job.numThreads;
		}
	}
	// Start threads -- ITERARE PARALLEL - This is the main amount of the computation (98%)
	for (size_t t=0; t<job.numThreads; t++) vThreads[t].Run(static_cast<int>(t));

	// Barrier
	bool bFinished=false;
	while (!bFinished) {
		bFinished = true;
		for (size_t t=0; t<job.numThreads; t++) bFinished = bFinished && vThreads[t].IsFinished();
		if (!bFinished) Sleep(20);
	}

	// Merge results -- SERIAL
	size_t LUTentries=job.LUT->size()/3;
	for (size_t y=0; y<job.height; y++) {
		for (size_t x=0; x<job.width; x++) {
			double d=static_cast<double>(job.scratch[2*(x+y*(job.width+1))  ] + 
										 job.scratch[2*(x+y*(job.width+1))+1] + 
										 job.scratch[2*((x+1)+y*(job.width+1))] + 
										 job.scratch[2*(x+(y+1)*(job.width+1))+1])/4.0;
			if (d==static_cast<double>(job.max_iter)) {
				for (int i=0; i<3; i++) job.rgb[3*(x+y*job.width)+i]=0;
			} 
			else {
				int id=static_cast<int>(d)%LUTentries;
				double w=d-floor(d);				
				for (int i=0; i<3; i++) {
					double c = static_cast<double>((*job.LUT)[3*id+i])+w*(static_cast<double>((*job.LUT)[3*((id+1)%LUTentries)+i])-static_cast<double>((*job.LUT)[3*id+i]));
					job.rgb[3*(x+y*job.width)+i]=static_cast<unsigned char>(max(min(c,255.0),0.0));
				}
			}
		}
	}
	// kill threads
	for (size_t t=0; t<job.numThreads; t++) {
		delete vThreads[t].J;
	}	
}



void Mandelbrot(const JOBDESC& job) {
	COMPLEX c;
	double dy=1.0/static_cast<double>(job.height)*(job.ur.i-job.ll.i);
	double dx=1.0/static_cast<double>(job.height)*(job.ur.i-job.ll.i);
	// Compute Quad flipped samples
	for (size_t y=0; y<job.height+1; y++) {		
		c.i=job.ll.i+static_cast<double>(y)*dy;
		for (size_t x=0; x<job.width+1; x++) {
			c.r=job.ll.r+static_cast<double>(x)*dx;
			MandelbrotIteration(&job.scratch[x+y*(job.width+1)],c,job);
		}
	}
	size_t LUTentries=job.LUT->size()/3;
	for (size_t y=0; y<job.height; y++) {
		for (size_t x=0; x<job.width; x++) {
			if (job.scratch[x+y*(job.width+1)]==job.max_iter) {
				for (int i=0; i<3; i++) job.rgb[3*(x+y*job.width)+i]=0;
			}
			else {
				double d=static_cast<double>(job.scratch[x+y*(job.width+1)]);
				int id=static_cast<int>(d)%LUTentries;
				double w=d-floor(d);
				for (int i=0; i<3; i++) {
					double c = static_cast<double>((*job.LUT)[3*id+i])+w*(static_cast<double>((*job.LUT)[3*((id+1)%LUTentries)+i])-static_cast<double>((*job.LUT)[3*id+i]));
					job.rgb[3*(x+y*job.width)+i]=static_cast<unsigned char>(max(min(c,255.0),0.0));
				}
			}
		}
	}
}

vector<unsigned char> LoadLUT(const char* name) {
	vector<unsigned char> LUT;
	FILE *fptr=NULL;
	errno_t error=fopen_s(&fptr,name,"r");
	if (error) return LUT;
	int r,g,b;
	while (!feof(fptr)) {
		fscanf_s(fptr,"%i %i %i\n",&r,&g,&b);
		LUT.push_back(static_cast<unsigned char>(r));
		LUT.push_back(static_cast<unsigned char>(g));
		LUT.push_back(static_cast<unsigned char>(b));
	}
	// replicate last entry (convenient for later interpolation)
	LUT.push_back(static_cast<unsigned char>(r));
	LUT.push_back(static_cast<unsigned char>(g));
	LUT.push_back(static_cast<unsigned char>(b));
	fclose(fptr);
	return LUT;
}

vector<unsigned char> GenerateDefaultLUT(size_t resolution) {
	const int numR=36;
	const unsigned char vKnotR[2*numR]={ // form: x,y in [0,255]^2
		  0,255,	 10,105,	 14, 64,	 18,155,
		 30,255,	 49,255,	 58, 64,	 59, 79,
		 61, 95,	 62, 96,	 63, 84,	 65, 38,
		 66,  0,	 94,  0,	100,254,	103,157,
		107,  0,	130,  0,	137,218,	138,240,
		140,228,	147,166,	148,147,	155, 52,
		156, 64,	166,255,	174,255,	177,226,
		178,255,	188,181,	210, 46,	211, 64,
		223,213,	228,241,	237,108,	251,206
	};
	const int numG=51;
	const unsigned char vKnotG[2*numG]={
		 14,  0,	 15, 10,	 17, 48,	 18, 79,
		 23,106,	 29,155,	 30,168,	 32,109,
		 33, 73,	 38,255,	 49,255,	 58, 64,
		 66,255,	 80, 64,	 83, 64,	 84, 91,
		 94,255,	100,254,	101,226,	104,177,
		106,161,	107,160,	111, 92,	112, 50,
		113, 19,	114,  0,	130,  0,	137,219,
		138,219,	140,192,	147,  0,	166,  0,
		174,220,	177,224,	178,232,	183,171,
		188,121,	198, 48,	207,  7,	211,  0,
		214, 14,	217, 40,	222,102,	223,123,
		224,131,	227,174,	228,199,	229,162,
		231,105,	235, 24,	237,  0
	};
	const int numB=30;
	const unsigned char vKnotB[2*numB]={
		 18,  0,	 24, 36,	 30, 92,	 33, 37,
		 34,110,	 38,255,     39,247,     49,  0,
		 80,  0,     84, 89,     93,234,     94,255,
		107,255,	111,159,	114, 64,	130,255,
		147,255,	156, 64,	166,255,	177,255,
		178,225,	183,159,	189, 98,	198, 38,
		203, 17,	211,  0,	215,  7,	223, 45,
		228, 17,	236,  0
	};
	vector<unsigned char> vResult;
	double w,c;
	unsigned char r,g,b;
	for (size_t s=0; s<resolution; s++) {
		double p=static_cast<double>(s)/static_cast<double>(resolution-1)*255.0;
		// RED
		c=w=0.0;
		for (size_t i=0; i<numR; i++) {
			double dist=static_cast<double>(vKnotR[2*i])-p;
			dist=pow(fabs(dist),5.0);
			dist=1.0/max(dist,1e-32);
			w+=dist;
			c+=static_cast<double>(vKnotR[2*i+1])*dist;
		}
		c/=w;
		r=static_cast<unsigned char>(max(min(c,255.0),0.0));
		vResult.push_back(r);
		// GREEN
		c=w=0.0;
		for (size_t i=0; i<numG; i++) {
			double dist=static_cast<double>(vKnotG[2*i])-p;
			dist=pow(fabs(dist),5.0);
			dist=1.0/max(dist,1e-32);
			w+=dist;
			c+=static_cast<double>(vKnotG[2*i+1])*dist;
		}
		c/=w;
		g=static_cast<unsigned char>(max(min(c,255.0),0.0));
		vResult.push_back(g);
		// BLUE
		c=w=0.0;
		for (size_t i=0; i<numB; i++) {
			double dist=static_cast<double>(vKnotB[2*i])-p;
			dist=pow(fabs(dist),5.0);
			dist=1.0/max(dist,1e-32);
			w+=dist;
			c+=static_cast<double>(vKnotB[2*i+1])*dist;
		}
		c/=w;
		b=static_cast<unsigned char>(max(min(c,255.0),0.0));
		vResult.push_back(b);
	}
	return vResult;
}

int main(int argc, char** argv) {
	size_t pageX=1024;
	size_t pageY=1024;
	size_t numPagesX = RESX;	// in 1K blocks
	size_t numPagesY = RESY;	// in 1K blocks

	// Create Large Array
	LargeArray2D<unsigned char,3> storage;
	unsigned int resx=static_cast<unsigned int>(numPagesX*pageX);
	unsigned int resy=static_cast<unsigned int>(numPagesY*pageY);
	wchar_t name[MAX_STRING_SIZE];
	swprintf_s(name,MAX_STRING_SIZE,OUTPUT,0);

#ifndef PYRAMID_ONLY
	storage.Create(resx,resy,pageX,pageY,name,MEM_USAGE*(1<<20));
	pageX = storage.GetPageSizeX();
	pageY = storage.GetPageSizeY();
	numPagesX = resx/pageX;
	numPagesY = resy/pageY;
	printf("Computing %ix%i pages, each with %ix%i pixels\n",numPagesX,numPagesY,pageX,pageY);
		
	// Compute Color LUT
	printf("Creating color LUT... ");
	vector<unsigned char> LUT=GenerateDefaultLUT(512);
	printf("done.\n");
	
	// The region of the fractal we want to see...
	COMPLEX ll(-2.1,-1.45);
	COMPLEX ur( 0.8, 1.45);

	// and the increment from one tile to the next
	COMPLEX db=COMPLEX((ur.r-ll.r)/static_cast<double>(numPagesX),(ur.i-ll.i)/static_cast<double>(numPagesY));

	// here's our job descriptor for one tile
	JOBDESC job;
	job.width    = pageX;
	job.height   = pageY;
	job.max_iter = 4096;
	job.escapeSq = 1e64; // escapeSq must be >= 4.0. The larger, the more shades will be used for the outer region.
	job.LUT      = &LUT;
	job.numThreads = 4;
	job.allocateBuffers();	// allocate enough space to store one page and an additional supersampling (scratch) buffer
	
	// And GO...
	CTimer ctTotal;
	for (size_t y=0; y<numPagesY; y++) {
		double di=static_cast<double>(y);
		for (size_t x=0; x<numPagesX; x++) {
			double dr=static_cast<double>(x);
			printf("Block %.4i,%.4i : ",x,y);
			job.ll = COMPLEX(ll.r+dr*db.r,ll.i+di*db.i);
			job.ur = job.ll+db;
			CTimer ct;
			ParallelMandelbrot(job);
			printf(" finished in %.2fs\n",static_cast<float>(ct.Query()));
			// move to Large Array
			storage.CopyFrom(job.rgb,x*pageX,y*pageY,pageX,pageY);
		}
	}
	printf("Took %.2fs total time\n",static_cast<float>(ctTotal.Query()));

	// Debugging: save as raw file
#ifdef TUM3D_WRITE_RAW_FILE
	storage.ConvertToRAW(L"mb.raw");
#endif

	// And done.
	job.freeBuffers();
	storage.Close();
#endif PYRAMID_ONLY

	// Now generate the mip-hierarchy
	int i=1;
	size_t targetResX=numPagesX*pageX/2;
	size_t targetResY=numPagesY*pageY/2;
	LargeArray2D<unsigned char,3> input;
	LargeArray2D<unsigned char,3> output;
	while (targetResX>0 && targetResY>0) {
		CTimer ctLevel;
		printf("Generating Level %i: %ix%i pixels\n",i,static_cast<int>(targetResX),static_cast<int>(targetResY));
		// Open input and output files, each has half of memory usage for caching
		wprintf(L"\n>> Input file: %s  [R]\n",name);
		printf(">> %ix%i pixels\n",targetResX*2,targetResY*2);
		input.Open(name,true,MEM_USAGE*(1<<19));
		swprintf_s(name,MAX_STRING_SIZE,OUTPUT,i);
		wprintf(L"\n>> Output file: %s [RW]\n",name);
		printf(">> %ix%i pixels\n",targetResX,targetResY);
		output.Create(targetResX,targetResY,pageX,pageY,name,MEM_USAGE*(1<<19),true);
		// Traverse all pages
		for (unsigned int pY=0; pY<max(1,output.GetSizeY()/output.GetPageSizeY()); pY++) {
			for (unsigned int pX=0; pX<max(1,output.GetSizeX()/output.GetPageSizeX()); pX++) {
				
				unsigned int yl=pY*output.GetPageSizeY();
				unsigned int yr=min((pY+1)*output.GetPageSizeY(),output.GetSizeY());

				unsigned int xl=pX*output.GetPageSizeX();
				unsigned int xr=min((pX+1)*output.GetPageSizeX(),output.GetSizeX());

				// Traverse all pixels in output page
				for (unsigned int y=yl; y<yr; y++) {
					unsigned int yInL=min(2*y,input.GetSizeY()-1);
					unsigned int yInR=min(2*y+1,input.GetSizeY()-1);
					for (unsigned int x=xl; x<xr; x++) {
						unsigned int xInL=min(2*x,input.GetSizeX()-1);
						unsigned int xInR=min(2*x+1,input.GetSizeX()-1);
						for (int c=0; c<3; c++) {
							float f=0.25f*(	static_cast<float>(input.Get(xInL,yInL)[c]) + 
											static_cast<float>(input.Get(xInR,yInL)[c]) + 
											static_cast<float>(input.Get(xInL,yInR)[c]) + 
											static_cast<float>(input.Get(xInR,yInR)[c]));
							output.Get(x,y)[c]=static_cast<unsigned char>(f);
						}
					}
				}
			}
		}
		input.Close();
		output.Close();
		printf("done [%.2fs]\n",static_cast<float>(ctLevel.Query()));
		i++;
		if (targetResX==1 && targetResY==1) {
			targetResX=0;
			targetResY=0;
		}
		else {
			targetResX=max((targetResX+1)/2,1);
			targetResY=max((targetResY+1)/2,1);
		}
	}


	return 0;
}
